# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import mlflow
import urllib3
import pytest

import client as cl
import merlin
from merlin.model import ModelVersion

from unittest import mock
from unittest.mock import patch, MagicMock

default_resource_request = cl.ResourceRequest(min_replica=1, max_replica=1, cpu_request="100m", memory_request="128Mi")
env_1 = cl.Environment(
    id=1,
    name="dev",
    cluster="cluster-1",
    is_default=True,
    default_resource_request=default_resource_request,
)
env_2 = cl.Environment(
    id=2,
    name="dev-2",
    cluster="cluster-2",
    is_default=False,
    default_resource_request=default_resource_request,
)


def test_set_url(url, use_google_oauth):
    merlin.set_url(url, use_google_oauth=use_google_oauth)
    assert url == merlin.get_url()


def test_set_project(url, project, use_google_oauth):
    with patch("urllib3.PoolManager.request") as mock_request:
        # expect exception when setting project but client is not set
        mock_request.return_value = _mock_get_project_call_empty_result()
        with pytest.raises(Exception):
            merlin.set_project(project.name)
            
        mock_request.return_value = _mock_get_project_call(project)

        merlin.set_url(url, use_google_oauth=use_google_oauth)
        merlin.set_project(project.name)

        assert merlin.active_project().name == project.name
        assert merlin.active_project().id == project.id
        assert merlin.active_project().mlflow_tracking_url == project.mlflow_tracking_url
        
def test_set_model(url, project, model, use_google_oauth):
    with patch("urllib3.PoolManager.request") as mock_request:
        # expect exception when setting model but client and project is not set
        mock_request.return_value = _mock_get_project_call_empty_result()
        with pytest.raises(Exception):
            merlin.set_model(model.name, model.type)

        merlin.set_url(url, use_google_oauth=use_google_oauth)

        with pytest.raises(Exception):
            merlin.set_model(model.name, model.type)
        
        mock_request.return_value = _mock_get_project_call(project)
        merlin.set_project(project.name)

        mock_request.side_effect = [_mock_get_project_call(project), _mock_get_model_call(project, model)]
        merlin.set_model(model.name, model.type)

        assert merlin.active_model().name == model.name
        assert merlin.active_model().type == model.type
        assert merlin.active_model().id == model.id
        assert merlin.active_model().mlflow_experiment_id == model.mlflow_experiment_id
        
def test_new_model_version(url, project, model, version, use_google_oauth):
    with patch("urllib3.PoolManager.request") as mock_request:
        # expect exception when creating new model  version but client and
        # project is not set
        mock_request.return_value = _mock_get_model_call_empty_result(project)
        with pytest.raises(Exception):
            with merlin.new_model_version() as v:
                print(v)

        merlin.set_url(url, use_google_oauth=use_google_oauth)

        with pytest.raises(Exception):
            with merlin.new_model_version() as v:
                print(v)

        mock_request.return_value = _mock_get_project_call(project)
        merlin.set_project(project.name)

        mock_request.return_value = _mock_get_model_call_empty_result(project)
        with pytest.raises(Exception):
            with merlin.new_model_version() as v:
                print(v)

        mock_request.side_effect = [_mock_get_project_call(project), _mock_get_model_call(project, model)]
        merlin.set_model(model.name, model.type)

        mock_request.side_effect = [_mock_get_project_call(project), _mock_get_model_call(project, model), _mock_new_model_version_call(model, version)]
        with merlin.new_model_version() as v:
            assert v is not None
            assert isinstance(v, ModelVersion)

            assert v.mlflow_run_id == version.mlflow_run_id
            
def test_new_model_version_with_labels(
    url, project, model, version, use_google_oauth
):
    with patch("urllib3.PoolManager.request") as mock_request:
        merlin.set_url(url, use_google_oauth=use_google_oauth)
        
        mock_request.return_value = _mock_get_project_call(project)
        merlin.set_project(project.name)
        
        mock_request.side_effect = [_mock_get_project_call(project), _mock_get_model_call(project, model)]
        merlin.set_model(model.name, model.type)

        # # Insert labels
        labels = {"model": "T-800", "software": "skynet"}
        mock_request.side_effect = [_mock_get_project_call(project), _mock_get_model_call(project, model), _mock_new_model_version_call(model, version, labels)]

        with merlin.new_model_version(labels=labels) as v:
            assert v is not None
            assert isinstance(v, ModelVersion)

            assert v.mlflow_run_id == version.mlflow_run_id
            for key, value in v.labels.items():
                assert labels[key] == value

def test_list_environment(url, use_google_oauth):
    with patch("urllib3.PoolManager.request") as mock_request:
        merlin.set_url(url, use_google_oauth=use_google_oauth)

        mock_request.return_value = _mock_list_environment_call()
        envs = merlin.list_environment()

        assert len(envs) == 2
        assert envs[0].name == env_1.name
        assert envs[1].name == env_2.name
        
def test_get_environment(url, use_google_oauth):
    with patch("urllib3.PoolManager.request") as mock_request:
        merlin.set_url(url, use_google_oauth=use_google_oauth)

        mock_request.return_value = _mock_list_environment_call()
        env = merlin.get_environment(env_1.name)
        assert env is not None
        assert env.name == env_1.name

        env = merlin.get_environment("undefined_env")
        assert env is None
        
def test_get_default_environment(url, use_google_oauth):
    with patch("urllib3.PoolManager.request") as mock_request:
        merlin.set_url(url, use_google_oauth=use_google_oauth)
        
        mock_request.return_value = _mock_list_environment_call()
        env = merlin.get_default_environment()

        assert env is not None
        assert env.name == env_1.name
        assert env.is_default
        
def test_mlflow_methods(url, project, model, version, use_google_oauth):
    with patch("urllib3.PoolManager.request") as mock_request:
        merlin.set_url(url, use_google_oauth=use_google_oauth)
        
        mock_request.return_value = _mock_get_project_call(project)
        merlin.set_project(project.name)
        
        mock_request.side_effect = [_mock_get_project_call(project), _mock_get_model_call(project, model)]
        merlin.set_model(model.name, model.type)
        
        mock_request.side_effect = [_mock_get_project_call(project), _mock_get_model_call(project, model), _mock_new_model_version_call(model, version)]
        with merlin.new_model_version() as v:
            merlin.log_metric("metric", 0.1)
            merlin.log_param("param", "value")
            merlin.set_tag("tag", "value")
        run_id = v.mlflow_run_id
        run = mlflow.get_run(run_id=run_id)

        assert run.data.metrics["metric"] == 0.1
        assert run.data.params["param"] == "value"
        assert run.data.tags["tag"] == "value"

def _mock_get_project_call_empty_result() -> MagicMock:
    mock_response = MagicMock()
    mock_response.method = "GET"
    mock_response.status = 200
    mock_response.path = "/v1/projects"
    mock_response.data = bytes("[]", 'utf-8')
    mock_response.headers = {
        'content-type': 'application/json',
        'charset': 'utf-8'
    }
    
    return mock_response

def _mock_get_project_call(project) -> MagicMock:
    mock_response = MagicMock()
    mock_response.method = "GET"
    mock_response.status = 200
    mock_response.path = "/v1/projects"
    mock_response.data = bytes(f"""[{{
                        "id": {project.id},
                        "name": "{project.name}",
                        "mlflow_tracking_url": "{project.mlflow_tracking_url}",
                        "created_at": "{project.created_at}",
                        "updated_at": "{project.updated_at}"
                      }}]""", 'utf-8')
    mock_response.headers = {
        'content-type': 'application/json',
        'charset': 'utf-8'
    }
    
    return mock_response

def _mock_get_model_call(project, model) -> MagicMock:
    mock_response = MagicMock()
    mock_response.method = "GET"
    mock_response.status = 200
    mock_response.path = f"/v1/projects/{project.id}/models"
    mock_response.data = bytes(f"""[{{
                        "id": {model.id},
                        "mlflow_experiment_id": {model.mlflow_experiment_id},
                        "name": "{model.name}",
                        "type": "{model.type.value}",
                        "mlflow_url": "{model.mlflow_url}",
                        "created_at": "2019-09-04T03:09:13.842Z",
                        "updated_at": "2019-09-04T03:09:13.842Z"
                      }}]""", 'utf-8')
    mock_response.headers = {
        'content-type': 'application/json',
        'charset': 'utf-8'
    }
    
    return mock_response

def _mock_get_model_call_empty_result(project) -> MagicMock:
    mock_response = MagicMock()
    mock_response.method = "GET"
    mock_response.status = 200
    mock_response.path = f"/v1/projects/{project.id}/models"
    mock_response.data = bytes("[]", 'utf-8')
    mock_response.headers = {
        'content-type': 'application/json',
        'charset': 'utf-8'
    }
    
    return mock_response

def _mock_new_model_version_call(model, version, labels=None) -> MagicMock:
    body = {
        "id": version.id,
        "mlflow_run_id": version.mlflow_run_id,
        "is_serving": False,
        "mlflow_url": version.mlflow_url,
        "created_at": "2019-09-04T03:09:13.842Z",
        "updated_at": "2019-09-04T03:09:13.842Z",
    }
    if labels is not None:
        body["labels"] = labels

    mock_response = MagicMock()
    mock_response.method = "POST"
    mock_response.status = 201
    mock_response.path = f"/v1/models/{model.id}/versions"
    mock_response.data = json.dumps(body).encode('utf-8')
    mock_response.headers = {
        'content-type': 'application/json',
        'charset': 'utf-8'
    }
    
    return mock_response

def _mock_list_environment_call() -> MagicMock:
    mock_response = MagicMock()
    mock_response.method = "GET"
    mock_response.status = 200
    mock_response.path =  "/v1/environments"
    mock_response.data = json.dumps([env_1.to_dict(), env_2.to_dict()]).encode('utf-8')
    mock_response.headers = {
        'content-type': 'application/json',
        'charset': 'utf-8'
    }
    
    return mock_response


import re
from merlin.batch.big_query_util import validate_text


def mc_valid_table_id(table_id: str) -> bool:
    """
    Validate MaxCompute table which satisfied this format project.schema.table

    Ref: 
    * https://www.alibabacloud.com/help/en/maxcompute/user-guide/manage-projects-in-the-new-maxcompute-console
    * https://www.alibabacloud.com/help/en/maxcompute/user-guide/table-operations#section-ymq-wfc-2bg


    :param table_id: table
    :return: boolean
    """
    components = table_id.split(".")
    if len(components) != 3:
        return False

    project_id = components[0]
    _ = components[1]
    table = components[2]
    # TODO: check regex for valid project and schema
    # for now, just validate the table name
    # validate project name
    if not validate_text(project_id, r'^[a-zA-Z][a-zA-Z0-9_]*$', 28, 3):
        return False
    if not validate_text(table, r'^[a-zA-Z][a-zA-Z0-9_]*$', 128):
        return False
    return True

def mc_valid_column(column_name: str) -> bool:
    """
    Validate MaxCompute column

    :param column_name: column name
    :return: boolean
    """
    column_name_max_length = 128
    pattern = r'^[a-zA-Z_]\w*$'
    return validate_text(column_name, pattern, column_name_max_length)

def mc_valid_columns(columns) -> bool:
    """
    Validate MaxCompute columns

    :param columns: columns
    :return: boolean
    """
    for column in columns:
        if not mc_valid_column(column):
            return False
    return True


# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
import types
from unittest.mock import MagicMock, patch

import client
import client as cl
import pytest
from merlin import AutoscalingPolicy, DeploymentMode, MetricsType
from merlin.autoscaling import (
    RAW_DEPLOYMENT_DEFAULT_AUTOSCALING_POLICY,
    SERVERLESS_DEFAULT_AUTOSCALING_POLICY,
)
from merlin.batch.config import PredictionJobConfig, ResultType
from merlin.batch.job import JobStatus
from merlin.batch.sink import BigQuerySink, SaveMode
from merlin.batch.source import BigQuerySource
from merlin.endpoint import VersionEndpoint
from merlin.model import ModelType
from merlin.model_schema import InferenceSchema, ModelSchema, RankingOutput, ValueType
from merlin.protocol import Protocol
from merlin.model_observability import ModelObservability
from urllib3_mock import Responses

default_resource_request = cl.ResourceRequest(
    min_replica=1, max_replica=1, cpu_request="100m", memory_request="128Mi"
)
gpu = cl.GPUConfig(
    name="nvidia-tesla-p4",
    values=["1", "4", "8"],
    resource_type="nvidia.com/gpu",
    node_selector={"cloud.google.com/gke-accelerator": "nvidia-tesla-p4"},
    tolerations=[
        cl.GPUToleration(
            key="caraml/nvidia-tesla-p4",
            operator="Equal",
            value="enabled",
            effect="NoSchedule",
        ),
        cl.GPUToleration(
            key="nvidia.com/gpu", operator="Equal", value="present", effect="NoSchedule"
        ),
    ],
)

env_1 = cl.Environment(
    id=1,
    name="dev",
    cluster="cluster-1",
    is_default=True,
    default_resource_request=default_resource_request,
)
env_2 = cl.Environment(
    id=2,
    name="dev-2",
    cluster="cluster-2",
    is_default=False,
    default_resource_request=default_resource_request,
)
env_3 = cl.Environment(
    id=2,
    name="dev-3",
    cluster="cluster-3",
    is_default=False,
    default_resource_request=default_resource_request,
    gpus=[gpu],
)

ep1 = cl.VersionEndpoint(
    id="1234",
    version_id=1,
    status="running",
    url="localhost/1",
    service_name="svc-1",
    environment_name=env_1.name,
    environment=env_1,
    monitoring_url="grafana.com",
)
ep2 = cl.VersionEndpoint(
    id="4567",
    version_id=1,
    status="running",
    url="localhost/1",
    service_name="svc-2",
    environment_name=env_2.name,
    environment=env_2,
    monitoring_url="grafana.com",
)
ep3 = cl.VersionEndpoint(
    id="1234",
    version_id=1,
    status="running",
    url="localhost/1",
    service_name="svc-1",
    environment_name=env_1.name,
    environment=env_1,
    monitoring_url="grafana.com",
    deployment_mode="raw_deployment",
)
ep4 = cl.VersionEndpoint(
    id="1234",
    version_id=1,
    status="running",
    url="localhost/1",
    service_name="svc-1",
    environment_name=env_1.name,
    environment=env_1,
    monitoring_url="grafana.com",
    autoscaling_policy=client.AutoscalingPolicy(
        metrics_type=cl.MetricsType.CPU_UTILIZATION, target_value=10
    ),
)
resource_request_with_gpu = cl.ResourceRequest(
    min_replica=1,
    max_replica=1,
    cpu_request="100m",
    memory_request="128Mi",
    gpu_name="nvidia-tesla-p4",
    gpu_request="1",
)
ep5 = cl.VersionEndpoint(
    id="789",
    version_id=1,
    status="running",
    url="localhost/1",
    service_name="svc-1",
    environment_name=env_3.name,
    environment=env_3,
    monitoring_url="grafana.com",
    resource_request=resource_request_with_gpu,
)
upi_ep = cl.VersionEndpoint(
    id="1234",
    version_id=1,
    status="running",
    url="localhost/1",
    service_name="svc-1",
    environment_name=env_1.name,
    environment=env_1,
    monitoring_url="grafana.com",
    protocol=cl.Protocol.UPI_V1,
)
observability_enabled_ep = cl.VersionEndpoint(
    id="7899",
    version_id=1,
    status="running",
    url="localhost/1",
    service_name="svc-1",
    environment_name=env_3.name,
    environment=env_3,
    monitoring_url="grafana.com",
    enable_model_observability=True,
)

more_granular_observability_cfg_ep = cl.VersionEndpoint(
    id="8000",
    version_id=1,
    status="running",
    url="localhost/1",
    service_name="svc-1",
    environment_name=env_3.name,
    environment=env_3,
    monitoring_url="grafana.com",
    model_observability=cl.ModelObservability(
        enabled=True,
        ground_truth_source=cl.GroundTruthSource(
            table_urn="table_urn",
            event_timestamp_column="event_timestamp_column",
            source_project="dwh_project",
        ),
        ground_truth_job=cl.GroundTruthJob(
            cron_schedule="cron_schedule",
            service_account_secret_name="service_account_secret_name",
            start_day_offset_from_now=1,
            end_day_offset_from_now=1,
            cpu_request="cpu_request",
            cpu_limit="cpu_limit",
            memory_request="memory_request",
            memory_limit="memory_limit",
            grace_period_day=1,
        ),
        prediction_log_ingestion_resource_request=cl.PredictionLogIngestionResourceRequest(
            replica=1, cpu_request="1", memory_request="1Gi"
        ),
    ),
)

rule_1 = cl.ModelEndpointRule(
    destinations=[
        cl.ModelEndpointRuleDestination(version_endpoint_id=ep1.id, weight=100)
    ]
)
rule_2 = cl.ModelEndpointRule(
    destinations=[
        cl.ModelEndpointRuleDestination(version_endpoint_id=ep2.id, weight=100)
    ]
)
mdl_endpoint_1 = cl.ModelEndpoint(
    id=1,
    model_id=1,
    model=None,
    status="serving",
    url="localhost/1",
    rule=rule_1,
    environment_name=env_1.name,
    environment=env_1,
)
mdl_endpoint_2 = cl.ModelEndpoint(
    id=2,
    model_id=1,
    model=None,
    status="serving",
    url="localhost/2",
    rule=rule_2,
    environment_name=env_2.name,
    environment=env_2,
)

mdl_endpoint_upi = cl.ModelEndpoint(
    id=1,
    model_id=1,
    model=None,
    status="serving",
    url="localhost/1",
    rule=rule_1,
    environment_name=env_1.name,
    environment=env_1,
    protocol=cl.Protocol.UPI_V1,
)

config = {
    "job_config": {
        "version": "v1",
        "kind": "PredictionJob",
        "name": "job-1",
        "bigquery_source": {
            "table": "project.dataset.source_table",
            "features": ["feature_1", "feature_2"],
            "options": {
                "key": "val",
            },
        },
        "model": {
            "type": "PYFUNC_V2",
            "uri": "gs://my-model/model",
            "result": {"type": "DOUBLE"},
            "options": {
                "key": "val",
            },
        },
        "bigquery_sink": {
            "table": "project.dataset.result_table",
            "staging_bucket": "gs://test",
            "result_column": "prediction",
            "save_mode": 1,
            "options": {
                "key": "val",
            },
        },
    },
    "image_ref": "asia.gcr.io/my-image:1",
    "service_account_name": "my-service-account",
    "resource_request": {
        "driver_cpu_request": "1",
        "driver_memory_request": "1Gi",
        "executor_cpu_request": "1",
        "executor_memory_request": "1Gi",
        "executor_replica": 1,
    },
}
job_1 = cl.PredictionJob(
    id=1,
    name="job-1",
    version_id=1,
    model_id=1,
    config=config,
    status="pending",
    error="",
    created_at="2019-08-29T08:13:12.377Z",
    updated_at="2019-08-29T08:13:12.377Z",
)
job_2 = cl.PredictionJob(
    id=2,
    name="job-2",
    version_id=1,
    model_id=1,
    config=config,
    status="pending",
    error="error",
    created_at="2019-08-29T08:13:12.377Z",
    updated_at="2019-08-29T08:13:12.377Z",
)


def serialize_datetime(obj):
    if isinstance(obj, datetime.datetime):
        return obj.isoformat()
    raise TypeError("Type is not serializable")


class TestProject:
    secret_1 = cl.Secret(id=1, name="secret-1", data="secret-data-1")
    secret_2 = cl.Secret(id=2, name="secret-2", data="secret-data-2")

    def test_create_secret(self, project):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response = MagicMock()
            mock_response.method = "POST"
            mock_response.status = 200
            mock_response.path = "/v1/projects/1/secrets"
            mock_response.data = json.dumps(self.secret_1.to_dict()).encode('utf-8')
            mock_response.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }

            mock_request.return_value = mock_response
   
            project.create_secret(self.secret_1.name, self.secret_1.data)
            
            _, last_call_kwargs = mock_request.call_args_list[-1]
            actual_body = json.loads(last_call_kwargs.get("body"))
            
            assert actual_body["name"] == self.secret_1.name
            assert actual_body["data"] == self.secret_1.data

    def test_update_secret(self, project):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/projects/1/secrets"
            mock_response_1.data = json.dumps([self.secret_1.to_dict(), self.secret_2.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "PATCH"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/projects/1/secrets/1"
            mock_response_2.data = json.dumps(self.secret_1.to_dict()).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2]

            project.update_secret(self.secret_1.name, "new-data")

            _, last_call_kwargs = mock_request.call_args_list[-1]
            actual_body = json.loads(last_call_kwargs.get("body"))
            
            assert actual_body["name"] == self.secret_1.name
            assert actual_body["data"] == "new-data"

            mock_response_3 = MagicMock()
            mock_response_3.method = "GET"
            mock_response_3.status = 200
            mock_response_3.path = "/v1/projects/1/secrets"
            mock_response_3.data = json.dumps([self.secret_1.to_dict()]).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            mock_request.side_effect = [mock_response_3]

            with pytest.raises(
                ValueError,
                match=f"unable to find secret {self.secret_2.name} in project {project.name}",
            ):
                project.update_secret(self.secret_2.name, "new-data")
                
    def test_delete_secret(self, project):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/projects/1/secrets"
            mock_response_1.data = json.dumps([self.secret_1.to_dict(), self.secret_2.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "DELETE"
            mock_response_2.status = 204
            mock_response_2.path = "/v1/projects/1/secrets/1"
            mock_response_2.data = json.dumps({}).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2]

            project.delete_secret(self.secret_1.name)

            mock_response_3 = MagicMock()
            mock_response_3.method = "GET"
            mock_response_3.status = 200
            mock_response_3.path = "/v1/projects/1/secrets"
            mock_response_3.data = json.dumps([self.secret_1.to_dict()]).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            mock_request.side_effect = [mock_response_3]

            with pytest.raises(
                ValueError,
                match=f"unable to find secret {self.secret_2.name} in project {project.name}",
            ):
                project.delete_secret(self.secret_2.name)
                
    def test_list_secret(self, project):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response = MagicMock()
            mock_response.method = "GET"
            mock_response.status = 200
            mock_response.path = "/v1/projects/1/secrets"
            mock_response.data = json.dumps([self.secret_1.to_dict(), self.secret_2.to_dict()]).encode('utf-8')
            mock_response.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }

            mock_request.return_value = mock_response

            secret_names = project.list_secret()
            assert secret_names == [self.secret_1.name, self.secret_2.name]
            
class TestModelVersion:
    def test_list_endpoint(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response = MagicMock()
            mock_response.method = "GET"
            mock_response.status = 200
            mock_response.path = "/v1/models/1/versions/1/endpoint"
            mock_response.data = json.dumps([ep1.to_dict(), ep2.to_dict()]).encode('utf-8')
            mock_response.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.return_value = mock_response

            endpoints = version.list_endpoint()
            assert len(endpoints) == 2
            assert endpoints[0].id == ep1.id
            assert endpoints[1].id == ep2.id
            
    def test_deploy(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/environments"
            mock_response_1.data = json.dumps([env_1.to_dict(), env_2.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions/1/endpoint"
            mock_response_2.data = json.dumps([]).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_3 = MagicMock()
            mock_response_3.method = "POST"
            mock_response_3.status = 201
            mock_response_3.path = "/v1/models/1/versions/1/endpoint"
            mock_response_3.data = json.dumps(ep1.to_dict()).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_4 = MagicMock()
            mock_response_4.method = "GET"
            mock_response_4.status = 200
            mock_response_4.path = "/v1/models/1/versions/1/endpoint/1234"
            mock_response_4.data = json.dumps(ep1.to_dict()).encode('utf-8')
            mock_response_4.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_5 = MagicMock()
            mock_response_5.method = "GET"
            mock_response_5.status = 200
            mock_response_5.path = "/v1/models/1/versions/1/endpoint"
            mock_response_5.data = json.dumps([ep1.to_dict()]).encode('utf-8')
            mock_response_5.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2, mock_response_3, mock_response_4, mock_response_5]

            endpoint = version.deploy(environment_name=env_1.name)

            assert endpoint.id == ep1.id
            assert endpoint.status.value == ep1.status
            assert endpoint.environment_name == ep1.environment_name
            assert endpoint.environment.cluster == env_1.cluster
            assert endpoint.environment.name == env_1.name
            assert endpoint.deployment_mode == DeploymentMode.SERVERLESS
            assert endpoint.autoscaling_policy == SERVERLESS_DEFAULT_AUTOSCALING_POLICY
            assert endpoint.protocol == Protocol.HTTP_JSON
            
    def test_deploy_upiv1(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/environments"
            mock_response_1.data = json.dumps([env_1.to_dict(), env_2.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions/1/endpoint"
            mock_response_2.data = json.dumps([]).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }

            mock_response_3 = MagicMock()
            mock_response_3.method = "POST"
            mock_response_3.status = 201
            mock_response_3.path = "/v1/models/1/versions/1/endpoint"
            mock_response_3.data = json.dumps(upi_ep.to_dict()).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
        
            mock_response_4 = MagicMock()
            mock_response_4.method = "GET"
            mock_response_4.status = 200
            mock_response_4.path = "/v1/models/1/versions/1/endpoint/1234"
            mock_response_4.data = json.dumps(upi_ep.to_dict()).encode('utf-8')
            mock_response_4.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
        
            mock_response_5 = MagicMock()
            mock_response_5.method = "GET"
            mock_response_5.status = 200
            mock_response_5.path = "/v1/models/1/versions/1/endpoint"
            mock_response_5.data = json.dumps([upi_ep.to_dict()]).encode('utf-8')
            mock_response_5.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2, mock_response_3, mock_response_4, mock_response_5]

            endpoint = version.deploy(environment_name=env_1.name)

            assert endpoint.id == upi_ep.id
            assert endpoint.status.value == upi_ep.status
            assert endpoint.environment_name == upi_ep.environment_name
            assert endpoint.environment.cluster == env_1.cluster
            assert endpoint.environment.name == env_1.name
            assert endpoint.deployment_mode == DeploymentMode.SERVERLESS
            assert endpoint.autoscaling_policy == SERVERLESS_DEFAULT_AUTOSCALING_POLICY
            assert endpoint.protocol == Protocol.UPI_V1
            
    def test_deploy_using_raw_deployment_mode(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/environments"
            mock_response_1.data = json.dumps([env_1.to_dict(), env_2.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions/1/endpoint"
            mock_response_2.data = json.dumps([]).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }

            mock_response_3 = MagicMock()
            mock_response_3.method = "POST"
            mock_response_3.status = 201
            mock_response_3.path = "/v1/models/1/versions/1/endpoint"
            mock_response_3.data = json.dumps(ep3.to_dict()).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
        
            mock_response_4 = MagicMock()
            mock_response_4.method = "GET"
            mock_response_4.status = 200
            mock_response_4.path = "/v1/models/1/versions/1/endpoint/1234"
            mock_response_4.data = json.dumps(ep3.to_dict()).encode('utf-8')
            mock_response_4.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
        
            mock_response_5 = MagicMock()
            mock_response_5.method = "GET"
            mock_response_5.status = 200
            mock_response_5.path = "/v1/models/1/versions/1/endpoint"
            mock_response_5.data = json.dumps([ep3.to_dict()]).encode('utf-8')
            mock_response_5.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2, mock_response_3, mock_response_4, mock_response_5]

            endpoint = version.deploy(
                environment_name=env_1.name, deployment_mode=DeploymentMode.RAW_DEPLOYMENT
            )

            assert endpoint.id == ep3.id
            assert endpoint.status.value == ep3.status
            assert endpoint.environment_name == ep3.environment_name
            assert endpoint.environment.cluster == env_1.cluster
            assert endpoint.environment.name == env_1.name
            assert endpoint.deployment_mode == DeploymentMode.RAW_DEPLOYMENT
            assert endpoint.autoscaling_policy == RAW_DEPLOYMENT_DEFAULT_AUTOSCALING_POLICY
            
    def test_deploy_with_autoscaling_policy(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/environments"
            mock_response_1.data = json.dumps([env_1.to_dict(), env_2.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions/1/endpoint"
            mock_response_2.data = json.dumps([]).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }

            mock_response_3 = MagicMock()
            mock_response_3.method = "POST"
            mock_response_3.status = 201
            mock_response_3.path = "/v1/models/1/versions/1/endpoint"
            mock_response_3.data = json.dumps(ep4.to_dict()).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
        
            mock_response_4 = MagicMock()
            mock_response_4.method = "GET"
            mock_response_4.status = 200
            mock_response_4.path = "/v1/models/1/versions/1/endpoint/1234"
            mock_response_4.data = json.dumps(ep4.to_dict()).encode('utf-8')
            mock_response_4.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
        
            mock_response_5 = MagicMock()
            mock_response_5.method = "GET"
            mock_response_5.status = 200
            mock_response_5.path = "/v1/models/1/versions/1/endpoint"
            mock_response_5.data = json.dumps([ep4.to_dict()]).encode('utf-8')
            mock_response_5.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2, mock_response_3, mock_response_4, mock_response_5]

            endpoint = version.deploy(
                environment_name=env_1.name,
                autoscaling_policy=AutoscalingPolicy(
                    metrics_type=MetricsType.CPU_UTILIZATION, target_value=10
                ),
            )

            assert endpoint.id == ep4.id
            assert endpoint.status.value == ep4.status
            assert endpoint.environment_name == ep4.environment_name
            assert endpoint.environment.cluster == env_1.cluster
            assert endpoint.environment.name == env_1.name
            assert endpoint.deployment_mode == DeploymentMode.SERVERLESS
            assert endpoint.autoscaling_policy.metrics_type == MetricsType.CPU_UTILIZATION
            assert endpoint.autoscaling_policy.target_value == 10
            
    def test_deploy_default_env(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/models/1/versions/1/endpoint"
            mock_response_1.data = json.dumps([]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/environments"
            mock_response_2.data = json.dumps([env_2.to_dict()]).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2]
            
            with pytest.raises(ValueError):
                version.deploy()
                
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/environments"
            mock_response_1.data = json.dumps([env_1.to_dict(), env_2.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions/1/endpoint"
            mock_response_2.data = json.dumps([]).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }

            mock_response_3 = MagicMock()
            mock_response_3.method = "POST"
            mock_response_3.status = 201
            mock_response_3.path = "/v1/models/1/versions/1/endpoint"
            mock_response_3.data = json.dumps(ep1.to_dict()).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
        
            mock_response_4 = MagicMock()
            mock_response_4.method = "GET"
            mock_response_4.status = 200
            mock_response_4.path = "/v1/models/1/versions/1/endpoint/1234"
            mock_response_4.data = json.dumps(ep1.to_dict()).encode('utf-8')
            mock_response_4.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
        
            mock_response_5 = MagicMock()
            mock_response_5.method = "GET"
            mock_response_5.status = 200
            mock_response_5.path = "/v1/models/1/versions/1/endpoint"
            mock_response_5.data = json.dumps([ep1.to_dict()]).encode('utf-8')
            mock_response_5.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2, mock_response_3, mock_response_4, mock_response_5]

            endpoint = version.deploy()

            assert endpoint.id == ep1.id
            assert endpoint.status.value == ep1.status
            assert endpoint.environment_name == ep1.environment_name
            assert endpoint.environment.cluster == env_1.cluster
            assert endpoint.environment.name == env_1.name
            
    def test_redeploy_model(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/environments"
            mock_response_1.data = json.dumps([env_1.to_dict(), env_2.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions/1/endpoint"
            mock_response_2.data = json.dumps([ep3.to_dict()]).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_3 = MagicMock()
            mock_response_3.method = "PUT"
            mock_response_3.status = 200
            mock_response_3.path = "/v1/models/1/versions/1/endpoint/1234"
            mock_response_3.data = json.dumps(ep4.to_dict()).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_4 = MagicMock()
            mock_response_4.method = "GET"
            mock_response_4.status = 200
            mock_response_4.path = "/v1/models/1/versions/1/endpoint/1234"
            mock_response_4.data = json.dumps(ep4.to_dict()).encode('utf-8')
            mock_response_4.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_5 = MagicMock()
            mock_response_5.method = "GET"
            mock_response_5.status = 200
            mock_response_5.path = "/v1/models/1/versions/1/endpoint"
            mock_response_5.data = json.dumps([ep4.to_dict()]).encode('utf-8')
            mock_response_5.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2, mock_response_3, mock_response_4, mock_response_5]

            # Redeployment (add autoscaling policy and change deployment mode)
            endpoint = version.deploy(
                environment_name=env_1.name,
                autoscaling_policy=AutoscalingPolicy(
                    metrics_type=MetricsType.CPU_UTILIZATION, target_value=10
                ),
            )

            assert endpoint.id == ep4.id
            assert endpoint.status.value == ep4.status
            assert endpoint.environment_name == ep4.environment_name
            assert endpoint.environment.cluster == env_1.cluster
            assert endpoint.environment.name == env_1.name
            assert endpoint.deployment_mode == DeploymentMode.SERVERLESS
            assert endpoint.autoscaling_policy.metrics_type == MetricsType.CPU_UTILIZATION
            assert endpoint.autoscaling_policy.target_value == 10
            
    def test_deploy_with_gpu(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/environments"
            mock_response_1.data = json.dumps([env_3.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions/1/endpoint"
            mock_response_2.data = json.dumps([]).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_3 = MagicMock()
            mock_response_3.method = "POST"
            mock_response_3.status = 201
            mock_response_3.path = "/v1/models/1/versions/1/endpoint"
            mock_response_3.data = json.dumps(ep5.to_dict()).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_4 = MagicMock()
            mock_response_4.method = "GET"
            mock_response_4.status = 200
            mock_response_4.path = "/v1/models/1/versions/1/endpoint/789"
            mock_response_4.data = json.dumps(ep5.to_dict()).encode('utf-8')
            mock_response_4.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_5 = MagicMock()
            mock_response_5.method = "GET"
            mock_response_5.status = 200
            mock_response_5.path = "/v1/models/1/versions/1/endpoint"
            mock_response_5.data = json.dumps([ep5.to_dict()]).encode('utf-8')
            mock_response_5.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2, mock_response_3, mock_response_4, mock_response_5]

            endpoint = version.deploy(environment_name=env_3.name)

            assert endpoint.id == ep5.id
            assert endpoint.status.value == ep5.status
            assert endpoint.environment_name == ep5.environment_name
            assert endpoint.environment.cluster == env_3.cluster
            assert endpoint.environment.name == env_3.name
            assert endpoint.deployment_mode == DeploymentMode.SERVERLESS
            assert endpoint.resource_request.gpu_name == resource_request_with_gpu.gpu_name
            assert (
                endpoint.resource_request.gpu_request
                == resource_request_with_gpu.gpu_request
            )
    
    def test_deploy_with_model_observability_enabled(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/environments"
            mock_response_1.data = json.dumps([env_3.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions/1/endpoint"
            mock_response_2.data = json.dumps([]).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_3 = MagicMock()
            mock_response_3.method = "POST"
            mock_response_3.status = 201
            mock_response_3.path = "/v1/models/1/versions/1/endpoint"
            mock_response_3.data = json.dumps(observability_enabled_ep.to_dict()).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_4 = MagicMock()
            mock_response_4.method = "GET"
            mock_response_4.status = 200
            mock_response_4.path = "/v1/models/1/versions/1/endpoint/7899"
            mock_response_4.data = json.dumps(observability_enabled_ep.to_dict()).encode('utf-8')
            mock_response_4.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_5 = MagicMock()
            mock_response_5.method = "GET"
            mock_response_5.status = 200
            mock_response_5.path = "/v1/models/1/versions/1/endpoint"
            mock_response_5.data = json.dumps([observability_enabled_ep.to_dict()]).encode('utf-8')
            mock_response_5.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2, mock_response_3, mock_response_4, mock_response_5]

            endpoint = version.deploy(
                environment_name=env_3.name, enable_model_observability=True
            )

            assert endpoint.id == observability_enabled_ep.id
            assert endpoint.status.value == observability_enabled_ep.status
            assert endpoint.environment_name == observability_enabled_ep.environment_name
            assert endpoint.environment.cluster == env_3.cluster
            assert endpoint.environment.name == env_3.name
            assert endpoint.deployment_mode == DeploymentMode.SERVERLESS
            assert endpoint.enable_model_observability == True
            
    def test_deploy_with_more_granular_model_observability_cfg(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/environments"
            mock_response_1.data = json.dumps([env_3.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions/1/endpoint"
            mock_response_2.data = json.dumps([]).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_3 = MagicMock()
            mock_response_3.method = "POST"
            mock_response_3.status = 201
            mock_response_3.path = "/v1/models/1/versions/1/endpoint"
            mock_response_3.data = json.dumps(more_granular_observability_cfg_ep.to_dict()).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_4 = MagicMock()
            mock_response_4.method = "GET"
            mock_response_4.status = 200
            mock_response_4.path = "/v1/models/1/versions/1/endpoint/8000"
            mock_response_4.data = json.dumps(more_granular_observability_cfg_ep.to_dict()).encode('utf-8')
            mock_response_4.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_5 = MagicMock()
            mock_response_5.method = "GET"
            mock_response_5.status = 200
            mock_response_5.path = "/v1/models/1/versions/1/endpoint"
            mock_response_5.data = json.dumps([more_granular_observability_cfg_ep.to_dict()]).encode('utf-8')
            mock_response_5.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2, mock_response_3, mock_response_4, mock_response_5]

            model_observability = ModelObservability.from_model_observability_response(more_granular_observability_cfg_ep.model_observability)
            endpoint = version.deploy(
                environment_name=env_3.name, model_observability=model_observability
            )

            assert endpoint.id == more_granular_observability_cfg_ep.id
            assert endpoint.status.value == more_granular_observability_cfg_ep.status
            assert endpoint.environment_name == more_granular_observability_cfg_ep.environment_name
            assert endpoint.environment.cluster == env_3.cluster
            assert endpoint.environment.name == env_3.name
            assert endpoint.deployment_mode == DeploymentMode.SERVERLESS
            assert endpoint.model_observability == model_observability

    def test_undeploy(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response = MagicMock()
            mock_response.method = "GET"
            mock_response.status = 200
            mock_response.path = "/v1/models/1/versions/1/endpoint"
            mock_response.data = json.dumps([ep2.to_dict()]).encode('utf-8')
            mock_response.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }

            mock_request.return_value = mock_response

            version.undeploy(environment_name=env_1.name)
            assert mock_request.call_count == 1
            
            mock_request.reset_mock()
            
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/models/1/versions/1/endpoint"
            mock_response_1.data = json.dumps([ep1.to_dict(), ep2.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "DELETE"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions/1/endpoint/1234"
            mock_response_2.data = json.dumps(ep1.to_dict()).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2]

            version.undeploy(environment_name=env_1.name)
            assert mock_request.call_count == 2
            
    def test_undeploy_default_env(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/models/1/versions/1/endpoint"
            mock_response_1.data = json.dumps([]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/environments"
            mock_response_2.data = json.dumps([env_2.to_dict()]).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2]
            with pytest.raises(ValueError):
                version.deploy()
                
            mock_request.reset_mock()

            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/environments"
            mock_response_1.data = json.dumps([env_1.to_dict(), env_2.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions/1/endpoint"
            mock_response_2.data = json.dumps([ep2.to_dict()]).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2]

            version.undeploy()
            assert mock_request.call_count == 2

            mock_request.reset_mock()
            
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/environments"
            mock_response_1.data = json.dumps([env_1.to_dict(), env_2.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions/1/endpoint"
            mock_response_2.data = json.dumps([ep1.to_dict(), ep2.to_dict()]).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_3 = MagicMock()
            mock_response_3.method = "DELETE"
            mock_response_3.status = 200
            mock_response_3.path = "/v1/models/1/versions/1/endpoint/1234"
            mock_response_3.data = json.dumps([ep1.to_dict(), ep2.to_dict()]).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }

            mock_request.side_effect = [mock_response_1, mock_response_2, mock_response_3]

            version.undeploy()
            assert mock_request.call_count == 3

    def test_list_prediction_job(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/models/1/versions/1/jobs-by-page?page=1"
            mock_response_1.data = json.dumps({
                "results": [job_1.to_dict()],
                "paging": {
                    "page": 1,
                    "pages": 2,
                    "total": 2,
                },
            }, default=serialize_datetime).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions/1/jobs-by-page?page=2"
            mock_response_2.data = json.dumps({
                "results": [job_2.to_dict()],
                "paging": {
                    "page": 2,
                    "pages": 2,
                    "total": 2,
                },
            }, default=serialize_datetime).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2]
        
            jobs = version.list_prediction_job()
            assert len(jobs) == 2
            assert jobs[0].id == job_1.id
            assert jobs[0].name == job_1.name
            assert jobs[0].status == JobStatus(job_1.status)
            assert jobs[0].error == job_1.error

            assert jobs[1].id == job_2.id
            assert jobs[1].name == job_2.name
            assert jobs[1].status == JobStatus(job_2.status)
            assert jobs[1].error == job_2.error
            
    def test_create_prediction_job(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            job_1.status = "completed"
            
            mock_response = MagicMock()
            mock_response.method = "POST"
            mock_response.status = 200
            mock_response.path = "/v1/models/1/versions/1/jobs"
            mock_response.data = json.dumps(job_1.to_dict(), default=serialize_datetime).encode('utf-8')
            mock_response.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }

            bq_src = BigQuerySource(
                table="project.dataset.source_table",
                features=["feature_1", "feature2"],
                options={"key": "val"},
            )

            bq_sink = BigQuerySink(
                table="project.dataset.result_table",
                result_column="prediction",
                save_mode=SaveMode.OVERWRITE,
                staging_bucket="gs://test",
                options={"key": "val"},
            )

            job_config = PredictionJobConfig(
                source=bq_src,
                sink=bq_sink,
                service_account_name="my-service-account",
                result_type=ResultType.INTEGER,
            )
            
            mock_request.return_value = mock_response

            j = version.create_prediction_job(job_config=job_config)
            assert j.status == JobStatus.COMPLETED
            assert j.id == job_1.id
            assert j.error == job_1.error
            assert j.name == job_1.name

            _, last_call_kwargs = mock_request.call_args_list[-1]
            actual_req = json.loads(last_call_kwargs.get("body"))

            assert actual_req["config"]["job_config"]["bigquery_source"] == bq_src.to_dict()
            assert actual_req["config"]["job_config"]["bigquery_sink"] == bq_sink.to_dict()
            assert (
                actual_req["config"]["job_config"]["model"]["result"]["type"]
                == ResultType.INTEGER.value
            )
            assert (
                actual_req["config"]["job_config"]["model"]["uri"]
                == f"{version.artifact_uri}/model"
            )
            assert (
                actual_req["config"]["job_config"]["model"]["type"]
                == ModelType.PYFUNC_V2.value.upper()
            )
            assert actual_req["config"]["service_account_name"] == "my-service-account"
            
    @patch("merlin.model.DEFAULT_PREDICTION_JOB_DELAY", 0)
    @patch("merlin.model.DEFAULT_PREDICTION_JOB_RETRY_DELAY", 0)
    def test_create_prediction_job_with_retry_failed(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            job_1.status = "pending"

            mock_response = MagicMock()
            mock_response.method = "POST"
            mock_response.status = 200
            mock_response.path = "/v1/models/1/versions/1/jobs"
            mock_response.data = json.dumps(job_1.to_dict(), default=serialize_datetime).encode('utf-8')
            mock_response.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_responses = [mock_response]

            for i in range(5):
                temp_mock_response = MagicMock()
                temp_mock_response.method = "GET"
                temp_mock_response.status = 200
                temp_mock_response.path = "/v1/models/1/versions/1/jobs/1"
                temp_mock_response.data = json.dumps(job_1.to_dict(), default=serialize_datetime).encode('utf-8')
                temp_mock_response.headers = {
                    'content-type': 'application/json',
                    'charset': 'utf-8'
                }
                
                mock_responses.append(temp_mock_response)

            bq_src = BigQuerySource(
                table="project.dataset.source_table",
                features=["feature_1", "feature2"],
                options={"key": "val"},
            )

            bq_sink = BigQuerySink(
                table="project.dataset.result_table",
                result_column="prediction",
                save_mode=SaveMode.OVERWRITE,
                staging_bucket="gs://test",
                options={"key": "val"},
            )

            job_config = PredictionJobConfig(
                source=bq_src,
                sink=bq_sink,
                service_account_name="my-service-account",
                result_type=ResultType.INTEGER,
            )

            mock_request.side_effect = mock_responses
            with pytest.raises(ValueError):
                j = version.create_prediction_job(job_config=job_config)
                assert j.id == job_1.id
                assert j.error == job_1.error
                assert j.name == job_1.name
                assert mock_request.call_count == 6

    @patch("merlin.model.DEFAULT_PREDICTION_JOB_DELAY", 0)
    @patch("merlin.model.DEFAULT_PREDICTION_JOB_RETRY_DELAY", 0)
    def test_create_prediction_job_with_retry_success(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            job_1.status = "pending"

            mock_response_1 = MagicMock()
            mock_response_1.method = "POST"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/models/1/versions/1/jobs"
            mock_response_1.data = json.dumps(job_1.to_dict(), default=serialize_datetime).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_responses = [mock_response_1]

            for i in range(4):
                temp_mock_response = MagicMock()
                temp_mock_response.method = "GET"
                temp_mock_response.status = 200
                temp_mock_response.path = "/v1/models/1/versions/1/jobs/1"
                temp_mock_response.data = json.dumps(job_1.to_dict(), default=serialize_datetime).encode('utf-8')
                temp_mock_response.headers = {
                    'content-type': 'application/json',
                    'charset': 'utf-8'
                }
                
                mock_responses.append(temp_mock_response)

            job_1.status = "completed"
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions/1/jobs/1"
            mock_response_2.data = json.dumps(job_1.to_dict(), default=serialize_datetime).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_responses.append(mock_response_2)

            bq_src = BigQuerySource(
                table="project.dataset.source_table",
                features=["feature_1", "feature2"],
                options={"key": "val"},
            )

            bq_sink = BigQuerySink(
                table="project.dataset.result_table",
                result_column="prediction",
                save_mode=SaveMode.OVERWRITE,
                staging_bucket="gs://test",
                options={"key": "val"},
            )

            job_config = PredictionJobConfig(
                source=bq_src,
                sink=bq_sink,
                service_account_name="my-service-account",
                result_type=ResultType.INTEGER,
            )

            mock_request.side_effect = mock_responses
            
            j = version.create_prediction_job(job_config=job_config)
            assert j.status == JobStatus.COMPLETED
            assert j.id == job_1.id
            assert j.error == job_1.error
            assert j.name == job_1.name

            _, last_call_kwargs = mock_request.call_args_list[0]
            actual_req = json.loads(last_call_kwargs.get("body"))
            
            assert actual_req["config"]["job_config"]["bigquery_source"] == bq_src.to_dict()
            assert actual_req["config"]["job_config"]["bigquery_sink"] == bq_sink.to_dict()
            assert (
                actual_req["config"]["job_config"]["model"]["result"]["type"]
                == ResultType.INTEGER.value
            )
            assert (
                actual_req["config"]["job_config"]["model"]["uri"]
                == f"{version.artifact_uri}/model"
            )
            assert (
                actual_req["config"]["job_config"]["model"]["type"]
                == ModelType.PYFUNC_V2.value.upper()
            )
            assert actual_req["config"]["service_account_name"] == "my-service-account"
            assert mock_request.call_count == 6
            
    @patch("merlin.model.DEFAULT_PREDICTION_JOB_DELAY", 0)
    @patch("merlin.model.DEFAULT_PREDICTION_JOB_RETRY_DELAY", 0)
    def test_create_prediction_job_with_retry_pending_then_failed(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            job_1.status = "pending"

            mock_response_1 = MagicMock()
            mock_response_1.method = "POST"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/models/1/versions/1/jobs"
            mock_response_1.data = json.dumps(job_1.to_dict(), default=serialize_datetime).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_responses = [mock_response_1]

            for i in range(4):
                temp_mock_response = MagicMock()
                temp_mock_response.method = "GET"
                temp_mock_response.status = 200
                temp_mock_response.path = "/v1/models/1/versions/1/jobs/1"
                temp_mock_response.data = json.dumps(job_1.to_dict(), default=serialize_datetime).encode('utf-8')
                temp_mock_response.headers = {
                    'content-type': 'application/json',
                    'charset': 'utf-8'
                }
                
                mock_responses.append(temp_mock_response)

            job_1.status = "failed"
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions/1/jobs/1"
            mock_response_2.data = json.dumps(job_1.to_dict(), default=serialize_datetime).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_responses.append(mock_response_2)

            bq_src = BigQuerySource(
                table="project.dataset.source_table",
                features=["feature_1", "feature2"],
                options={"key": "val"},
            )
    
            bq_sink = BigQuerySink(
                table="project.dataset.result_table",
                result_column="prediction",
                save_mode=SaveMode.OVERWRITE,
                staging_bucket="gs://test",
                options={"key": "val"},
            )
    
            job_config = PredictionJobConfig(
                source=bq_src,
                sink=bq_sink,
                service_account_name="my-service-account",
                result_type=ResultType.INTEGER,
            )
    
            mock_request.side_effect = mock_responses
            with pytest.raises(ValueError):
                j = version.create_prediction_job(job_config=job_config)
                assert j.id == job_1.id
                assert j.error == job_1.error
                assert j.name == job_1.name
    
            assert mock_request.call_count == 6
            
    def test_stop_prediction_job(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            job_1.status = "pending"

            mock_response_1 = MagicMock()
            mock_response_1.method = "POST"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/models/1/versions/1/jobs"
            mock_response_1.data = json.dumps(job_1.to_dict(), default=serialize_datetime).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions/1/jobs/1"
            mock_response_2.data = json.dumps(job_1.to_dict(), default=serialize_datetime).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_3 = MagicMock()
            mock_response_3.method = "PUT"
            mock_response_3.status = 204
            mock_response_3.path = "/v1/models/1/versions/1/jobs/1/stop"
            mock_response_3.data = json.dumps({}).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }

            job_1.status = "terminated"
            
            mock_response_4 = MagicMock()
            mock_response_4.method = "GET"
            mock_response_4.status = 200
            mock_response_4.path = "/v1/models/1/versions/1/jobs/1"
            mock_response_4.data = json.dumps(job_1.to_dict(), default=serialize_datetime).encode('utf-8')
            mock_response_4.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }

            bq_src = BigQuerySource(
                table="project.dataset.source_table",
                features=["feature_1", "feature2"],
                options={"key": "val"},
            )

            bq_sink = BigQuerySink(
                table="project.dataset.result_table",
                result_column="prediction",
                save_mode=SaveMode.OVERWRITE,
                staging_bucket="gs://test",
                options={"key": "val"},
            )

            job_config = PredictionJobConfig(
                source=bq_src,
                sink=bq_sink,
                service_account_name="my-service-account",
                result_type=ResultType.INTEGER,
            )
            
            mock_request.side_effect = [mock_response_1, mock_response_2, mock_response_3, mock_response_4]

            j = version.create_prediction_job(job_config=job_config, sync=False)
            j = j.stop()
            assert j.status == JobStatus.TERMINATED
            assert j.id == job_1.id
            assert j.error == job_1.error
            assert j.name == job_1.name
            
    def test_model_version_deletion(self, version):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response = MagicMock()
            mock_response.method = "DELETE"
            mock_response.status = 200
            mock_response.path = "/v1/models/1/versions/1"
            mock_response.data = json.dumps(1).encode('utf-8')
            mock_response.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.return_value = mock_response
            
            response = version.delete_model_version()
            assert response == 1

class TestModel:
    v1 = cl.Version(id=1, model_id=1)
    v2 = cl.Version(id=2, model_id=1)
    v3 = cl.Version(
        id=3, model_id=1, labels={"model": "T-800"}, python_version="3.10.*"
    )

    schema = client.ModelSchema(
        id=3,
        model_id=1,
        spec=client.SchemaSpec(
            tag_columns=["tags", "extras"],
            feature_types={
                "featureA": client.ValueType.FLOAT64,
                "featureB": client.ValueType.INT64,
                "featureC": client.ValueType.BOOLEAN,
                "featureD": client.ValueType.STRING,
            },
            model_prediction_output=client.ModelPredictionOutput(
                client.RankingOutput(
                    rank_score_column="score",
                    relevance_score_column="relevance_score",
                    output_class=client.ModelPredictionOutputClass.RANKINGOUTPUT,
                )
            ),
        ),
    )
    merlin_model_schema = ModelSchema(
        id=3,
        model_id=1,
        spec=InferenceSchema(
            tag_columns=["tags", "extras"],
            feature_types={
                "featureA": ValueType.FLOAT64,
                "featureB": ValueType.INT64,
                "featureC": ValueType.BOOLEAN,
                "featureD": ValueType.STRING,
            },
            model_prediction_output=RankingOutput(
                rank_score_column="score", relevance_score_column="relevance_score"
            ),
        ),
    )
    v4 = cl.Version(
        id=4,
        model_id=1,
        labels={"model": "T-800"},
        python_version="3.10.*",
        model_schema=schema,
    )

    def test_list_version(self, model):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/models/1/versions?limit=50&cursor=&search="
            mock_response_1.data = json.dumps([self.v1.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8',
                "Next-Cursor": "abcdef"
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/versions?limit=50&cursor=abcdef&search="
            mock_response_2.data = json.dumps([self.v2.to_dict()]).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2]
            
            versions = model.list_version()
            assert len(versions) == 2
            assert versions[0].id == 1
            assert versions[1].id == 2
            
    def test_list_version_with_labels(self, model):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response = MagicMock()
            mock_response.method = "GET"
            mock_response.status = 200
            mock_response.path = "/v1/models/1/versions?limit=50&cursor=&search=labels%3Amodel+in+%28T-800%29"
            mock_response.data = json.dumps([self.v3.to_dict()]).encode('utf-8')
            mock_response.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.return_value = mock_response
        
            versions = model.list_version({"model": ["T-800"]})
            assert len(versions) == 1
            assert versions[0].id == 3
            assert versions[0].labels["model"] == "T-800"
            
    def test_list_endpoint(self, model):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response = MagicMock()
            mock_response.method = "GET"
            mock_response.status = 200
            mock_response.path = "/v1/models/1/endpoints"
            mock_response.data = json.dumps([mdl_endpoint_1.to_dict(), mdl_endpoint_2.to_dict()]).encode('utf-8')
            mock_response.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.return_value = mock_response

            endpoints = model.list_endpoint()
            assert len(endpoints) == 2
            assert endpoints[0].id == mdl_endpoint_1.id
            assert endpoints[1].id == mdl_endpoint_2.id
            
    def test_new_model_version(self, model):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response = MagicMock()
            mock_response.method = "POST"
            mock_response.status = 201
            mock_response.path = "/v1/models/1/versions"
            mock_response.data = json.dumps(self.v4.to_dict()).encode('utf-8')
            mock_response.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.return_value = mock_response

            mv = model.new_model_version(
                labels={"model": "T-800"}, model_schema=self.merlin_model_schema
            )
            assert mv._python_version == "3.10.*"
            assert mv._id == 4
            assert mv._model._id == 1
            assert mv._labels == {"model": "T-800"}
            assert mv._model_schema == self.merlin_model_schema
            
    def test_serve_traffic(self, model):
        with patch("urllib3.PoolManager.request") as mock_request:
            ve = VersionEndpoint(ep1)
            with pytest.raises(ValueError):
                model.serve_traffic([ve], environment_name=env_1.name)

            with pytest.raises(ValueError):
                model.serve_traffic({ve: -1}, environment_name=env_1.name)

            with pytest.raises(ValueError):
                model.serve_traffic({ve: 101}, environment_name=env_1.name)

            with pytest.raises(ValueError):
                model.serve_traffic(
                    {VersionEndpoint(ep2): 100}, environment_name=env_1.name
                )
                
            # test create
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/models/1/endpoints"
            mock_response_1.data = json.dumps([]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "POST"
            mock_response_2.status = 201
            mock_response_2.path = "/v1/models/1/endpoints"
            mock_response_2.data = json.dumps(mdl_endpoint_1.to_dict()).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            print("aku mock_response_2",mock_response_2)

            mock_request.side_effect = [mock_response_1, mock_response_2]
            
            endpoint = model.serve_traffic({ve: 100}, environment_name=env_1.name)
            assert endpoint.id == mdl_endpoint_1.id
            assert (
                endpoint.environment_name == env_1.name == mdl_endpoint_1.environment_name
            )
            assert endpoint.protocol == Protocol.HTTP_JSON
    
            # test update
    
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/models/1/endpoints"
            mock_response_1.data = json.dumps([mdl_endpoint_1.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/endpoints/1"
            mock_response_2.data = json.dumps(mdl_endpoint_1.to_dict()).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_3 = MagicMock()
            mock_response_3.method = "PUT"
            mock_response_3.status = 200
            mock_response_3.path = "/v1/models/1/endpoints/1"
            mock_response_3.data = json.dumps(mdl_endpoint_1.to_dict()).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2, mock_response_3]
            
            endpoint = model.serve_traffic({ve: 100}, environment_name=env_1.name)
            assert endpoint.id == mdl_endpoint_1.id
            assert (
                endpoint.environment_name == env_1.name == mdl_endpoint_1.environment_name
            )
            
    def test_stop_serving_traffic(self, model):
        with patch("urllib3.PoolManager.request") as mock_request:
            ve = VersionEndpoint(ep1)
            with pytest.raises(ValueError):
                model.serve_traffic([ve], environment_name=env_1.name)

            with pytest.raises(ValueError):
                model.serve_traffic({ve: -1}, environment_name=env_1.name)

            with pytest.raises(ValueError):
                model.serve_traffic({ve: 101}, environment_name=env_1.name)

            with pytest.raises(ValueError):
                model.serve_traffic(
                    {VersionEndpoint(ep2): 100}, environment_name=env_1.name
                )
            
            
            # test create
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/models/1/endpoints"
            mock_response_1.data = json.dumps([]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "POST"
            mock_response_2.status = 201
            mock_response_2.path = "/v1/models/1/endpoints"
            mock_response_2.data = json.dumps(mdl_endpoint_1.to_dict()).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2]
        
            endpoint = model.serve_traffic({ve: 100}, environment_name=env_1.name)
            assert endpoint.id == mdl_endpoint_1.id
            assert (
                endpoint.environment_name == env_1.name == mdl_endpoint_1.environment_name
            )
            
            mock_request.reset_mock()

            # test DELETE
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/models/1/endpoints"
            mock_response_1.data = json.dumps([mdl_endpoint_1.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/endpoints/1"
            mock_response_2.data = json.dumps(mdl_endpoint_1.to_dict()).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_3 = MagicMock()
            mock_response_3.method = "DELETE"
            mock_response_3.status = 200
            mock_response_3.path = "/v1/models/1/endpoints/1"
            mock_response_3.data = json.dumps({}).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2, mock_response_3]
            
            model.stop_serving_traffic(endpoint.environment_name)
            assert mock_request.call_count == 2

    def test_serve_traffic_default_env(self, model):
        with patch("urllib3.PoolManager.request") as mock_request:
            ve = VersionEndpoint(ep1)
            
            # no default environment
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/environments"
            mock_response_1.data = json.dumps([env_2.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.return_value = mock_response_1        

            with pytest.raises(ValueError):
                model.serve_traffic({ve: 100})

            # test create
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/environments"
            mock_response_1.data = json.dumps([env_1.to_dict(), env_2.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/endpoints"
            mock_response_2.data = json.dumps([]).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_3 = MagicMock()
            mock_response_3.method = "POST"
            mock_response_3.status = 201
            mock_response_3.path = "/v1/models/1/endpoints"
            mock_response_3.data = json.dumps(mdl_endpoint_1.to_dict()).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
        
            mock_request.side_effect = [mock_response_1, mock_response_2, mock_response_3]
            
            endpoint = model.serve_traffic({ve: 100})
            assert endpoint.id == mdl_endpoint_1.id
            assert (
                endpoint.environment_name == env_1.name == mdl_endpoint_1.environment_name
            )


            # test update
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/environments"
            mock_response_1.data = json.dumps([env_1.to_dict(), env_2.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/endpoints"
            mock_response_2.data = json.dumps([mdl_endpoint_1.to_dict()]).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_3 = MagicMock()
            mock_response_3.method = "GET"
            mock_response_3.status = 200
            mock_response_3.path = "/v1/models/1/endpoints/1"
            mock_response_3.data = json.dumps(mdl_endpoint_1.to_dict()).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_4 = MagicMock()
            mock_response_4.method = "PUT"
            mock_response_4.status = 200
            mock_response_4.path = "/v1/models/1/endpoints/1"
            mock_response_4.data = json.dumps(mdl_endpoint_1.to_dict()).encode('utf-8')
            mock_response_4.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2, mock_response_3, mock_response_4]
        
            endpoint = model.serve_traffic({ve: 100})
            assert endpoint.id == mdl_endpoint_1.id
            assert (
                endpoint.environment_name == env_1.name == mdl_endpoint_1.environment_name
            )
            
    def test_serve_traffic_upi(self, model):
        with patch("urllib3.PoolManager.request") as mock_request:
            ve = VersionEndpoint(upi_ep)
            # test create
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/models/1/endpoints"
            mock_response_1.data = json.dumps([]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "POST"
            mock_response_2.status = 201
            mock_response_2.path = "/v1/models/1/endpoints"
            mock_response_2.data = json.dumps(mdl_endpoint_upi.to_dict()).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2]
        
            endpoint = model.serve_traffic({ve: 100}, environment_name=env_1.name)
            assert endpoint.id == mdl_endpoint_upi.id
            assert (
                endpoint.environment_name == env_1.name == mdl_endpoint_1.environment_name
            )
            assert endpoint.protocol == Protocol.UPI_V1

            # test update
            mock_response_1 = MagicMock()
            mock_response_1.method = "GET"
            mock_response_1.status = 200
            mock_response_1.path = "/v1/models/1/endpoints"
            mock_response_1.data = json.dumps([mdl_endpoint_upi.to_dict()]).encode('utf-8')
            mock_response_1.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_2 = MagicMock()
            mock_response_2.method = "GET"
            mock_response_2.status = 200
            mock_response_2.path = "/v1/models/1/endpoints/1"
            mock_response_2.data = json.dumps(mdl_endpoint_upi.to_dict()).encode('utf-8')
            mock_response_2.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_response_3 = MagicMock()
            mock_response_3.method = "PUT"
            mock_response_3.status = 200
            mock_response_3.path = "/v1/models/1/endpoints/1"
            mock_response_3.data = json.dumps(mdl_endpoint_upi.to_dict()).encode('utf-8')
            mock_response_3.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }
            
            mock_request.side_effect = [mock_response_1, mock_response_2, mock_response_3]
        
            endpoint = model.serve_traffic({ve: 100}, environment_name=env_1.name)
            assert endpoint.id == mdl_endpoint_upi.id
            assert (
                endpoint.environment_name == env_1.name == mdl_endpoint_upi.environment_name
            )
            assert endpoint.protocol == Protocol.UPI_V1
            
    def test_model_deletion(self, model):
        with patch("urllib3.PoolManager.request") as mock_request:
            mock_response = MagicMock()
            mock_response.method = "DELETE"
            mock_response.status = 200
            mock_response.path = "/v1/projects/1/models/1"
            mock_response.data = json.dumps(1).encode('utf-8')
            mock_response.headers = {
                'content-type': 'application/json',
                'charset': 'utf-8'
            }

            mock_request.return_value = mock_response
            response = model.delete_model()
    
            assert response == 1


# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import client


class MountedMLPSecret:
    """
    The MLP secret mounted to the Merlin model/transformer as an environment variable.
    """

    def __init__(
        self,
        mlp_secret_name: str,
        env_var_name: str,
    ):
        self._mlp_secret_name = mlp_secret_name
        self._env_var_name = env_var_name

    @classmethod
    def from_response(cls, response: client.MountedMLPSecret) -> MountedMLPSecret:
        return MountedMLPSecret(
            mlp_secret_name=response.mlp_secret_name,
            env_var_name=response.env_var_name
        )

    @property
    def mlp_secret_name(self) -> str:
        return self._mlp_secret_name

    @mlp_secret_name.setter
    def mlp_secret_name(self, mlp_secret_name):
        self._mlp_secret_name = mlp_secret_name

    @property
    def env_var_name(self) -> str:
        return self._env_var_name

    @env_var_name.setter
    def env_var_name(self, env_var_name):
        self._env_var_name = env_var_name



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
import urllib3
import pytest

from sys import version_info
from unittest import mock
from unittest.mock import patch, MagicMock

import client as cl
from client import ApiClient, Configuration
from merlin.client import MerlinClient
from merlin.endpoint import Status
from merlin.model import Model, ModelType, Project
from merlin.util import guess_mlp_ui_url
from merlin.version import VERSION


@pytest.fixture
def mock_url():
    return "http://merlin.dev/api"


@pytest.fixture
def api_client(mock_url):
    config = Configuration()
    config.host = mock_url + "/v1"
    return ApiClient(config)


mlflow_tracking_url = "http://mlflow.api.merlin.dev"
created_at = "2019-08-29T08:13:12.377Z"
updated_at = "2019-08-29T08:13:12.377Z"

default_resource_request = cl.ResourceRequest(min_replica=1, max_replica=1, cpu_request="100m", memory_request="128Mi")
env_1 = cl.Environment(
    id=1,
    name="dev",
    cluster="cluster-1",
    is_default=True,
    default_resource_request=default_resource_request,
)
env_2 = cl.Environment(
    id=2,
    name="dev-2",
    cluster="cluster-2",
    is_default=False,
    default_resource_request=default_resource_request,
)
mdl_endpoint_1 = cl.ModelEndpoint(
    id=1,
    model_id=1,
    model=None,
    status="serving",
    url="localhost",
    rule=None,
    environment_name=env_1.name,
    environment=env_1,
    created_at=created_at,
    updated_at=updated_at,
)


def serialize_datetime(obj):
    if isinstance(obj, datetime.datetime):
        return obj.isoformat()
    raise TypeError("Type is not serializable")

def test_get_project(mock_url, use_google_oauth):
    with patch("urllib3.PoolManager.request") as mock_request:
        mock_response = MagicMock()
        mock_response.status = 200
        mock_response.data = bytes(f"""[{{
                            "id": 0,
                            "name": "my-project",
                            "mlflow_tracking_url": "http://mlflow.api.merlin.dev",
                            "created_at": "{created_at}",
                            "updated_at": "{updated_at}"
                          }}]""", 'utf-8')
        mock_response.headers = {
            'content-type': 'application/json',
            'charset': 'utf-8'
        }
        mock_request.return_value = mock_response
    
        m = MerlinClient(mock_url, use_google_oauth=use_google_oauth)
        p = m.get_project("my-project")
        
        args, _ = mock_request.call_args
        assert args[0] == "GET"
        assert args[1] == "http://merlin.dev/api/v1/projects?name=my-project"
    
        assert p.id == 0
        assert p.name == "my-project"
        assert p.mlflow_tracking_url == "http://mlflow.api.merlin.dev"
        assert p.url == mock_url
        assert isinstance(p.created_at, datetime.datetime)
        assert isinstance(p.updated_at, datetime.datetime)
        
def test_create_invalid_project_name(
    mock_url, use_google_oauth
):
    project_name = "invalidProjectName"

    client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)

    # Try to create project with invalid name. It must be fail
    with pytest.raises(Exception):
        assert client.get_project(project_name)
        
def test_create_model(mock_url, api_client, use_google_oauth):
    project_id = 1010
    mlflow_experiment_id = 1
    model_name = "my-model"
    project_name = "my-project"
    model_type = ModelType.XGBOOST
    mlflow_url = "http://mlflow.api.merlin.dev"

    mock_response_1 = MagicMock()
    mock_response_1.status = 200
    mock_response_1.path = f"/api/v1/projects/{project_id}/models"
    mock_response_1.data = bytes("[]", 'utf-8')
    mock_response_1.headers = {
        'content-type': 'application/json',
        'charset': 'utf-8'
    }
    
    mock_response_2 = MagicMock()
    mock_response_2.status = 201
    mock_response_2.path = f"/api/v1/projects/{project_id}/models"
    mock_response_2.headers = {
        'content-type': 'application/json',
        'charset': 'utf-8'
    }
    mock_response_2.data = bytes(f"""{{
                        "id": 0,
                        "project_id": {project_id},
                        "mlflow_experiment_id": {mlflow_experiment_id},
                        "name": "{model_name}",
                        "type": "{model_type.value}",
                        "mlflow_url": "{mlflow_url}",
                        "endpoints": [],
                        "created_at": "{created_at}",
                        "updated_at": "{updated_at}"
                      }}""", 'utf-8')

    with patch("urllib3.PoolManager.request") as mock_request:
        mock_request.side_effect = [mock_response_1, mock_response_2]
        
        client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)
        prj = cl.Project(
           id=project_id, 
           name=project_name, 
           mlflow_tracking_url=mlflow_tracking_url, 
           created_at=created_at, 
           updated_at=updated_at
        )
        project = Project(prj, mock_url, api_client)
        with mock.patch.object(client, "get_project", return_value=project):
            model = client.get_or_create_model(
                "my-model", project_name=project_name, model_type=model_type
            )
            
            _, last_call_kwargs = mock_request.call_args_list[-1]

            assert json.loads(last_call_kwargs.get("body")) == json.loads(
                f"""
            {{
                "name" : "{model_name}",
                "type" : "{model_type.value}"
            }}
            """
            )
            assert model.id == 0
            assert model.mlflow_experiment_id == mlflow_experiment_id
            assert model.name == model_name
            assert model.type == model_type
            assert model.mlflow_url == mlflow_tracking_url
            assert model.mlflow_experiment_id == mlflow_experiment_id
            assert isinstance(model.created_at, datetime.datetime)
            assert isinstance(model.updated_at, datetime.datetime)
            assert model.project == project
            assert (
                f"merlin-sdk/{VERSION}" in last_call_kwargs.get("headers", {}).get("User-Agent")
            )
            assert (
                f"python/{version_info.major}.{version_info.minor}.{version_info.micro}"
                in last_call_kwargs.get("headers", {}).get("User-Agent")
            )
            
def test_create_invalid_model_name(mock_url, use_google_oauth):
    model_name = "invalidModelName"
    project_name = "my-project"
    model_type = ModelType.XGBOOST

    client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)

    # Try to create model with invalid name. It must be fail
    with pytest.raises(Exception):
        assert client.get_or_create_model(model_name, project_name, model_type)
        
def test_get_model(mock_url, api_client, use_google_oauth):
    project_id = 1010
    mlflow_experiment_id = 1
    model_name = "my-model"
    project_name = "my-project"
    model_type = ModelType.XGBOOST
    mlflow_url = "http://mlflow.api.merlin.dev"

    mock_response_1 = MagicMock()
    mock_response_1.method = "GET"
    mock_response_1.status = 200
    mock_response_1.path = f"/api/v1/projects/{project_id}/models"
    mock_response_1.data = bytes(f"""[{{
                        "id": 1,
                        "project_id": {project_id},
                        "mlflow_experiment_id": {mlflow_experiment_id},
                        "name": "{model_name}",
                        "type": "{model_type.value}",
                        "mlflow_url": "{mlflow_url}",
                        "endpoints": [],
                        "created_at": "{created_at}",
                        "updated_at": "{updated_at}"
                      }}]""", 'utf-8')
    mock_response_1.headers = {
        'content-type': 'application/json',
        'charset': 'utf-8'
    }
    
    mock_response_2 = MagicMock()
    mock_response_2.method = "GET"
    mock_response_2.status = 200
    mock_response_2.path = f"/api/v1/models/1/endpoints"
    mock_response_2.data = json.dumps([mdl_endpoint_1.to_dict()], default=serialize_datetime).encode('utf-8')
    mock_response_2.headers = {
        'content-type': 'application/json',
        'charset': 'utf-8'
    }
    
    with patch("urllib3.PoolManager.request") as mock_request:
        mock_request.side_effect = [mock_response_1, mock_response_2]
        
        client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)
        prj = cl.Project(
            id=project_id, 
            name=project_name, 
            mlflow_tracking_url=mlflow_tracking_url, 
            created_at=created_at, 
            updated_at=updated_at
        )
        project = Project(prj, mock_url, api_client)
        with mock.patch.object(client, "get_project", return_value=project):
            model = client.get_or_create_model(
                "my-model", project_name=project_name, model_type=model_type
            )
            assert model.id == 1
            assert model.name == model_name
            assert model.type == model_type
            assert model.mlflow_url == mlflow_tracking_url
            assert model.mlflow_experiment_id == mlflow_experiment_id
            assert isinstance(model.created_at, datetime.datetime)
            assert isinstance(model.updated_at, datetime.datetime)
            assert model.project == project

            default_model_endpoint = model.endpoint
            assert default_model_endpoint is not None
            assert default_model_endpoint.status == Status.SERVING
            assert default_model_endpoint.environment_name == env_1.name
            
def test_new_model_version(mock_url, api_client, use_google_oauth):
    project_id = 1
    model_id = 1
    version_id = 2
    model_name = "my-model"
    project_name = "my-project"
    mlflow_experiment_id = 1
    mlflow_run_id = "c5c3b6b220b34c7496de8c0400b7c793"
    model_type = ModelType.TENSORFLOW
    mlflow_url = "http://mlflow.api.merlin.dev"
    artifact_uri = "gs://zltest/model"
    created_at = "2019-09-04T03:09:13.842Z"
    updated_at = "2019-09-04T03:09:13.843Z"

    with patch("urllib3.PoolManager.request") as mock_request:
        mock_response = MagicMock()
        mock_response.method = "POST"
        mock_response.status = 201
        mock_response.path = f"/api/v1/models/{model_id}/versions"
        mock_response.data = bytes(f"""{{
                                "id": {version_id},
                                "model_id": {model_id},
                                "mlflow_run_id": "{mlflow_run_id}",
                                "mlflow_url": "{mlflow_url}",
                                "artifact_uri": "{artifact_uri}",
                                "endpoints": [],
                                "mlflow_url": "{mlflow_url}",
                                "created_at": "{created_at}",
                                "updated_at": "{updated_at}"
                            }}""", 'utf-8')
        mock_response.headers = {
            'content-type': 'application/json',
            'charset': 'utf-8'
        }
        
        mock_request.return_value = mock_response

        client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)
        prj = cl.Project(
            id=project_id, 
            name=project_name, 
            mlflow_tracking_url=mlflow_tracking_url, 
            created_at=created_at, 
            updated_at=updated_at
        )
        project = Project(prj, mock_url, api_client)
        mdl = cl.Model(
            id=model_id,
            project_id=project_id,
            mlflow_experiment_id=mlflow_experiment_id,
            name=model_name,
            type=model_type.value,
            mlflow_url=mlflow_url,
            endpoints=None,
            created_at=created_at,
            updated_at=updated_at,
        )
        mdl = Model(mdl, project, api_client)
        with mock.patch.object(client, "get_model", return_value=mdl):
            mv = client.new_model_version(model_name, project_name)

            assert mv.id == version_id
            assert mv.mlflow_run_id == mlflow_run_id
            assert mv.mlflow_url == mlflow_url
            assert mv.properties is None
            assert isinstance(mv.created_at, datetime.datetime)
            assert isinstance(mv.updated_at, datetime.datetime)
            assert mv.model == mdl
            ui_url = guess_mlp_ui_url(mock_url)
            assert mv.url == f"{ui_url}/projects/1/models/{model_id}/versions"
            
def test_list_environments(mock_url, use_google_oauth):
    with patch("urllib3.PoolManager.request") as mock_request:
        mock_response = MagicMock()
        mock_response.method = "GET"
        mock_response.status = 200
        mock_response.path = "/api/v1/environments"
        mock_response.data = json.dumps([env_1.to_dict(), env_2.to_dict()]).encode('utf-8')
        mock_response.headers = {
            'content-type': 'application/json',
            'charset': 'utf-8'
        }
        
        mock_request.return_value = mock_response
        
        client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)
        envs = client.list_environment()

        assert len(envs) == 2
        assert envs[0].name == env_1.name
        assert envs[0].cluster == env_1.cluster
        assert envs[0].is_default == env_1.is_default
        assert envs[1].name == env_2.name
        assert envs[1].cluster == env_2.cluster
        assert envs[1].is_default == env_2.is_default

def test_get_environment(mock_url, use_google_oauth):
    with patch("urllib3.PoolManager.request") as mock_request:
        mock_response = MagicMock()
        mock_response.method = "GET"
        mock_response.status = 200
        mock_response.path = "/api/v1/environments"
        mock_response.data = json.dumps([env_1.to_dict(), env_2.to_dict()]).encode('utf-8')
        mock_response.headers = {
            'content-type': 'application/json',
            'charset': 'utf-8'
        }
        
        mock_request.return_value = mock_response

        client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)
        env = client.get_environment(env_1.name)

        assert env is not None
        assert env.name == env_1.name
        assert env.cluster == env_1.cluster
        assert env.is_default == env_1.is_default

        env = client.get_environment("undefined_env")
        assert env is None

def test_get_default_environment(mock_url, use_google_oauth):
    with patch("urllib3.PoolManager.request") as mock_request:
        mock_response = MagicMock()
        mock_response.method = "GET"
        mock_response.status = 200
        mock_response.path = "/api/v1/environments"
        mock_response.data = json.dumps([env_1.to_dict(), env_2.to_dict()]).encode('utf-8')
        mock_response.headers = {
            'content-type': 'application/json',
            'charset': 'utf-8'
        }
        
        mock_request.return_value = mock_response
        
        client = MerlinClient(mock_url, use_google_oauth=use_google_oauth)
        env = client.get_default_environment()

        assert env.name == env_1.name
        assert env.cluster == env_1.cluster
        assert env.is_default == env_1.is_default

        mock_response_2 = MagicMock()
        mock_response_2.method = "GET"
        mock_response_2.status = 200
        mock_response_2.path = "/api/v1/environments"
        mock_response_2.data = json.dumps([env_2.to_dict()]).encode('utf-8')
        mock_response_2.headers = {
            'content-type': 'application/json',
            'charset': 'utf-8'
        }

        mock_request.return_value = mock_response_2
        
        env = client.get_default_environment()

        assert env is None



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities

    The version of the OpenAPI document: 0.14.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Any, ClassVar, Dict, List, Optional, Union
from pydantic import BaseModel
from client.models.model_prediction_config import ModelPredictionConfig
from client.models.protocol import Protocol
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

class StandardTransformerSimulationRequest(BaseModel):
    """
    StandardTransformerSimulationRequest
    """ # noqa: E501
    payload: Optional[Union[str, Any]] = None
    headers: Optional[Union[str, Any]] = None
    config: Optional[Union[str, Any]] = None
    model_prediction_config: Optional[ModelPredictionConfig] = None
    protocol: Optional[Protocol] = None
    __properties: ClassVar[List[str]] = ["payload", "headers", "config", "model_prediction_config", "protocol"]

    model_config = {
        "populate_by_name": True,
        "validate_assignment": True
    }


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of StandardTransformerSimulationRequest from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        _dict = self.model_dump(
            by_alias=True,
            exclude={
            },
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of model_prediction_config
        if self.model_prediction_config:
            _dict['model_prediction_config'] = self.model_prediction_config.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Dict) -> Self:
        """Create an instance of StandardTransformerSimulationRequest from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "payload": obj.get("payload"),
            "headers": obj.get("headers"),
            "config": obj.get("config"),
            "model_prediction_config": ModelPredictionConfig.from_dict(obj.get("model_prediction_config")) if obj.get("model_prediction_config") is not None else None,
            "protocol": obj.get("protocol")
        })
        return _obj


